package toy.dialog;

import java.awt.*;
import java.awt.event.*;

import toy.*;
import toy.image.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyDebugPanel extends ToyRefreshablePanel
                           implements ActionListener, ItemListener

{
    public static final Font MONOSPACED_FONT = new Font("Monospaced", Font.PLAIN, 12);
    public static final Font SANSSERIF_FONT = new Font("SansSerif", Font.PLAIN, 12);
    public static final Font LABEL_FONT = new Font("SansSerif", Font.PLAIN, 14);


    private ToyWorkspace workspace;
    private ToyDebugHash debugHash;

    private ToyFocusManager focusManager;
    private ToyImageManager imageManager;

    private CheckboxMenuItem uninitializedErrorCheckbox;
    private CheckboxMenuItem overflowErrorCheckbox;
    private CheckboxMenuItem outOfBoundsErrorCheckbox;
    private CheckboxMenuItem closeExecutionCheckbox;

    private MenuItem stepMenuItem;
    private MenuItem runMenuItem;
    private MenuItem resetMenuItem;

    private Panel leftPanel;
        private Label outputLabel;
        private TextArea outputTextArea;

        private Panel buttonPanel;
            private Button stepButton;
            private Button runButton;
            private Button resetButton;

        private Label coreLabel;
        private TextArea coreTextArea;


    private Panel rightPanel;
        private Label inputLabel;
        private ToyInputPanel inputPanel;

        private Label programLabel;
        private List programList;

        private Label converterLabel;
        private ToyConverterPanel converterPanel;


    ////////////////////////////////////////////////////////////////////////////
    // the constructor
    public ToyDebugPanel(ToyFocusManager newFocusManager,
                         ToyImageManager newImageManager,
                         CheckboxMenuItem newUninitializedErrorCheckbox,
                         CheckboxMenuItem newOverflowErrorCheckbox,
                         CheckboxMenuItem newOutOfBoundsErrorCheckbox,
                         CheckboxMenuItem newCloseExecutionCheckbox,
                         MenuItem newStepMenuItem,
                         MenuItem newRunMenuItem,
                         MenuItem newResetMenuItem
                         )
    {
        super(new GridBagLayout());

        focusManager = newFocusManager;
        imageManager = newImageManager;

        uninitializedErrorCheckbox = newUninitializedErrorCheckbox;
            uninitializedErrorCheckbox.addItemListener(this);
        overflowErrorCheckbox = newOverflowErrorCheckbox;
        outOfBoundsErrorCheckbox = newOutOfBoundsErrorCheckbox;
        closeExecutionCheckbox = newCloseExecutionCheckbox;

        stepMenuItem = newStepMenuItem;
            stepMenuItem.addActionListener(this);
        runMenuItem = newRunMenuItem;
            runMenuItem.addActionListener(this);
        resetMenuItem = newResetMenuItem;
            resetMenuItem.addActionListener(this);

        setBackground(new Color(80, 100, 120));

        leftPanel = new Panel(new GridBagLayout());
        leftPanel.setBackground(getBackground());
        add(
            leftPanel,
            ToyToolkit.createGridBagConstraints(
                0, 0,
                1, 1,
                2.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5),
                0, 0
                )
            );
        {
            outputLabel = new Label();
            outputLabel.setFont(LABEL_FONT);
            leftPanel.add(
                outputLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 0,
                    2, 1,
                    3.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
            outputTextArea = new TextArea(25, 80);
            outputTextArea.addFocusListener(focusManager);
            outputTextArea.setBackground(new Color(192, 192, 192));
            outputTextArea.setEditable(false);
            outputTextArea.setFont(MONOSPACED_FONT);
            outputTextArea.setForeground(new Color(0, 0, 0));
            leftPanel.add(
                outputTextArea,
                ToyToolkit.createGridBagConstraints(
                    0, 1,
                    2, 1,
                    3.0, 60.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );


            buttonPanel = new Panel(new GridLayout(1, 3, 5, 0));
            buttonPanel.setBackground(getBackground());
            leftPanel.add(
                buttonPanel,
                ToyToolkit.createGridBagConstraints(
                    1, 2,
                    1, 1,
                    1.5, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(5, 5, 5, 0),
                    0, 0
                    )
                );
            {
                stepButton = new Button("Step");
                stepButton.addActionListener(this);
                stepButton.addFocusListener(focusManager);
                stepButton.setBackground(new Color(192, 192, 192));
                stepButton.setForeground(new Color(0, 0, 0));
                buttonPanel.add(stepButton);

                runButton = new Button("Run");
                runButton.addActionListener(this);
                runButton.addFocusListener(focusManager);
                runButton.setBackground(new Color(192, 192, 192));
                runButton.setForeground(new Color(0, 0, 0));
                buttonPanel.add(runButton);

                resetButton = new Button("Reset");
                resetButton.addActionListener(this);
                resetButton.addFocusListener(focusManager);
                resetButton.setBackground(new Color(192, 192, 192));
                resetButton.setForeground(new Color(0, 0, 0));
                buttonPanel.add(resetButton);
            }

            coreLabel = new Label("Detailed Core Dump:");
            coreLabel.setFont(LABEL_FONT);
            leftPanel.add(
                coreLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 2,
                    1, 1,
                    1.5, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
            coreTextArea = new TextArea(8, 80);
            coreTextArea.addFocusListener(focusManager);
            coreTextArea.setBackground(new Color(192, 192, 192));
            coreTextArea.setEditable(false);
            coreTextArea.setFont(MONOSPACED_FONT);
            coreTextArea.setForeground(new Color(0, 0, 0));
            leftPanel.add(
                coreTextArea,
                ToyToolkit.createGridBagConstraints(
                    0, 3,
                    2, 1,
                    3.0, 40.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
        }

        rightPanel = new Panel(new GridBagLayout());
        rightPanel.setBackground(getBackground());
        add(
            rightPanel,
            ToyToolkit.createGridBagConstraints(
                1, 0,
                1, 1,
                1.0, 1.0,
                GridBagConstraints.CENTER,
                GridBagConstraints.BOTH,
                new Insets(5, 5, 5, 5),
                0, 0
                )
            );
        {
            inputLabel = new Label("Stdin:");
            inputLabel.setFont(LABEL_FONT);
            rightPanel.add(
                inputLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 0,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
            inputPanel = new ToyInputPanel(focusManager);
            rightPanel.add(
                inputPanel,
                ToyToolkit.createGridBagConstraints(
                    0, 1,
                    1, 1,
                    1.0, 10.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            programLabel = new Label("Program Summary:");
            programLabel.setFont(LABEL_FONT);
            rightPanel.add(
                programLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 2,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
            programList = new List(10);
            programList.addFocusListener(focusManager);
            programList.addItemListener(this);
            programList.setBackground(new Color(192, 192, 192));
            programList.setFont(MONOSPACED_FONT);
            programList.setForeground(new Color(0, 0, 0));
            rightPanel.add(
                programList,
                ToyToolkit.createGridBagConstraints(
                    0, 3,
                    1, 1,
                    1.0, 60.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );

            converterLabel = new Label("16-bit Base Converter:");
            converterLabel.setFont(LABEL_FONT);
            rightPanel.add(
                converterLabel,
                ToyToolkit.createGridBagConstraints(
                    0, 4,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.SOUTHWEST,
                    GridBagConstraints.HORIZONTAL,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
            converterPanel = new ToyConverterPanel(focusManager);
            rightPanel.add(
                converterPanel,
                ToyToolkit.createGridBagConstraints(
                    0, 5,
                    1, 1,
                    1.0, 1.0,
                    GridBagConstraints.CENTER,
                    GridBagConstraints.BOTH,
                    new Insets(0, 0, 0, 0),
                    0, 0
                    )
                );
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // setWorkspace() allows the frame to change the workspace displayed by this
    // panel
    public void setWorkspace(ToyWorkspace newWorkspace)
    {
        workspace = newWorkspace;

        workspace.reset();

        inputPanel.setVirtualMachine(workspace.getVirtualMachine());

        stepButton.setEnabled(true);
        stepMenuItem.setEnabled(true);
        refresh();
    }

    ////////////////////////////////////////////////////////////////////////////
    // intercept setBounds() calls
    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        super.setBounds(x, y, width, height);

        doLayout();
    }

    ////////////////////////////////////////////////////////////////////////////
    // refresh() is the method called when the programList needs to be rebuilt
    // or the right line needs to be selected
    public void refresh()
    {
        if (!isVisible())
            return;

        int line;
        ToyDebugHash debugHash = workspace.getDebugHash();

        inputPanel.refresh();

        stepButton.setEnabled(!workspace.getVirtualMachine().isDone());
        stepMenuItem.setEnabled(!workspace.getVirtualMachine().isDone());
        runButton.setEnabled(!workspace.getVirtualMachine().isDone());
        runMenuItem.setEnabled(!workspace.getVirtualMachine().isDone());

        //update the label
        if (workspace.getFileName() == null ||
            workspace.getFileName().length() == 0)
            outputLabel.setText(workspace.getName() + " Output:");
        else
            outputLabel.setText(workspace.getName() + " Output: (" +
                workspace.getFileName() + ")");


        //redo the programList
        //temporarily remove this item while we perform a ton of operations
        programList.setVisible(false);

        programList.removeAll();

        for (int ctr = 0; ctr < debugHash.getCardinality(); ctr++)
            programList.add(debugHash.getCode(ctr));

        //highlight the like where programCtr points except when we've halted
        //normally
        if (!workspace.getVirtualMachine().isDone() ||
            workspace.getVirtualMachine().hasEncounteredError())
        {
            line = debugHash.getLine(
                workspace.getVirtualMachine().getProgramCtr()
                );

            //if the line is not undefined, then select it
            //otherwise select nothing
            if (line != -1)
            {
                programList.select(Math.max(0, line - 1));
                programList.select(Math.min(debugHash.getCardinality() - 1, line + 1));

                programList.select(line);
            }
        }

        //ok we're done, add it back to the panel
        programList.setVisible(true);

        //redo the coreTextArea
        coreTextArea.setText(
            workspace.getVirtualMachine().getDetailedCoreDump(
                uninitializedErrorCheckbox.getState()
                )
            );

        //redo the outputTextArea
        int selectionStart = 0, selectionEnd = 0;
        ToyWord stdout[] = workspace.getVirtualMachine().getStdout();
        String outputText =
            "Core dump before running.\n" +
            "-------------------------\n" +
            workspace.getInitialVirtualMachine().getCoreDump(
                uninitializedErrorCheckbox.getState()
                ) + "\n" +
            "TOY console.\n" +
            "-------------------------\n";

        for (int ctr = 0; ctr < stdout.length; ctr++)
            outputText += stdout[ctr].toHexString() + "  " + stdout[ctr] + "\n";

        if (workspace.getVirtualMachine().hasEncounteredError())
        {
            selectionStart = outputText.length();
            outputText += "********************************************************************************\n";
            outputText += workspace.getVirtualMachine().getStderr();
            outputText += "********************************************************************************\n";
            selectionEnd = outputText.length();
        }

        if (workspace.getVirtualMachine().isDone())
        {
            outputText += "\n\n";
            outputText += "Core dump after running.\n";
            outputText += "------------------------\n";
            outputText += workspace.getVirtualMachine().getCoreDump(
                    uninitializedErrorCheckbox.getState()
                    );
        }

        outputTextArea.setText(outputText);
        if (workspace.getVirtualMachine().hasEncounteredError())
        {
            outputTextArea.setSelectionStart(selectionStart);
            outputTextArea.setSelectionEnd(selectionEnd);
        }
        else
            outputTextArea.setCaretPosition(outputText.length());
    }


    ////////////////////////////////////////////////////////////////////////////
    // Implement ActionListener
    public void actionPerformed(ActionEvent e)
    {
        if (e.getSource() == stepButton ||
            e.getSource() == stepMenuItem && isVisible())
        {
            workspace.getVirtualMachine().step(
                uninitializedErrorCheckbox.getState(),
                overflowErrorCheckbox.getState(),
                outOfBoundsErrorCheckbox.getState()
                );

            refresh();

            if (workspace.getVirtualMachine().requiresInput())
            {
                ToyPopup.askForInput(this, workspace.getVirtualMachine());

                //make the inputPanel update itself with the newly entered data
                //(as opposed to the other way around)
                inputPanel.setVirtualMachine(workspace.getVirtualMachine());
            }
        }

        if (e.getSource() == runButton ||
            e.getSource() == runMenuItem && isVisible())
        {
            refresh();

            ToyPopup.execute(
                this,
                this,
                imageManager,
                workspace.getVirtualMachine(),
                uninitializedErrorCheckbox.getState(),
                overflowErrorCheckbox.getState(),
                outOfBoundsErrorCheckbox.getState(),
                closeExecutionCheckbox
                );

            refresh();
        }

        if (e.getSource() == resetButton ||
            e.getSource() == resetMenuItem && isVisible())
        {
            workspace.reset();

            refresh();
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ItemListener
    public void itemStateChanged(ItemEvent e)
    {
        if (e.getSource() == programList)
        {
            //highlight the like where programCtr points except when we've halted
            //normally
            if (!workspace.getVirtualMachine().isDone() ||
                workspace.getVirtualMachine().hasEncounteredError())
            {
                debugHash = workspace.getDebugHash();

                int line = debugHash.getLine(
                    workspace.getVirtualMachine().getProgramCtr()
                    );

                //if the line is not undefined, then select it
                //otherwise select nothing
                if (line != -1)
                {
                    programList.select(Math.max(0, line - 1));
                    programList.select(Math.min(debugHash.getCardinality() - 1, line + 1));

                    programList.select(line);
                }
            }
        }

        if (e.getSource() == uninitializedErrorCheckbox && isVisible())
            refresh();
    }
}