package toy.dialog;

import java.awt.*;
import java.awt.event.*;

import toy.image.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyImageButton extends Component
{
    private int disabledImageIndex, imageIndex, hoverImageIndex, activeImageIndex;
    private ToyImageManager imageManager;

    private ActionListener actionListener;

    private boolean mouseInside;
    private boolean mouseDown;
    private boolean hasFocus;


    public ToyImageButton(ToyImageManager newImageManager,
                          int newDisabledImageIndex, int newImageIndex,
                          int newHoverImageIndex, int newActiveImageIndex)
    {
        imageManager = newImageManager;

        disabledImageIndex = newDisabledImageIndex;
        imageIndex = newImageIndex;
        hoverImageIndex = newHoverImageIndex;
        activeImageIndex = newActiveImageIndex;

        enableEvents(AWTEvent.FOCUS_EVENT_MASK);
        enableEvents(AWTEvent.KEY_EVENT_MASK);
        enableEvents(AWTEvent.MOUSE_EVENT_MASK);
        enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);
    }



    ////////////////////////////////////////////////////////////////////////////
    // Describe our new Component

    public boolean isFocusTraversable()
    {
        return isEnabled();
    }

    public boolean isDoubleBuffered()
    {
        return false;
    }

    public boolean isLightweight()
    {
        return true;
    }

    ////////////////////////////////////////////////////////////////////////////
    // The get and set functions

    public boolean hasFocus()
    {
        return hasFocus;
    }

    public void setEnabled(boolean b)
    {
        super.setEnabled(b);
        repaint();
    }

    public void setVisible(boolean b)
    {
        if (b == isVisible())
            return;

        super.setVisible(b);

        if (getParent() != null)
        {
            if (b)
                repaint();
            else
                getParent().repaint(
                    getBounds().x,
                    getBounds().y,
                    getBounds().width,
                    getBounds().height
                    );
        }
    }

    public void setImageIndices(int newDisabledImageIndex, int newImageIndex,
                                int newHoverImageIndex, int newActiveImageIndex)
    {
        disabledImageIndex = newDisabledImageIndex;
        imageIndex = newImageIndex;
        hoverImageIndex = newHoverImageIndex;
        activeImageIndex = newActiveImageIndex;

        repaint();
    }

    public int getDisabledImageIndex()
    {
        return disabledImageIndex;
    }

    public int getImageIndex()
    {
        return imageIndex;
    }

    public int getHoverImageIndex()
    {
        return hoverImageIndex;
    }

    public int getActiveImageIndex()
    {
        return activeImageIndex;
    }

    ////////////////////////////////////////////////////////////////////////////
    // Allow ActionListeners to work

    public void addActionListener(ActionListener newActionListener)
    {
        actionListener = AWTEventMulticaster.add(actionListener, newActionListener);
    }


    public void removeActionListener(ActionListener newActionListener)
    {
       actionListener = AWTEventMulticaster.remove(actionListener, newActionListener);
    }


    ////////////////////////////////////////////////////////////////////////////
    // Overriding the painting functions
    public void repaint()
    {
        paint(getGraphics());
    }

    public void update(Graphics g)
    {
        paint(g);
    }

    public void paint(Graphics g)
    {
        if (getParent() == null || !isVisible() || !getParent().isVisible() || g == null)
            return;

        if (!isEnabled())
            g.drawImage(
                imageManager.getImage(disabledImageIndex),
                0,
                0,
                this
                );
        else {
            if (mouseDown && mouseInside)
                g.drawImage(
                    imageManager.getImage(activeImageIndex),
                    0,
                    0,
                    this
                    );
            else {
                if (mouseInside || hasFocus)
                    g.drawImage(
                        imageManager.getImage(hoverImageIndex),
                        0,
                        0,
                        this
                        );
                else
                    g.drawImage(
                        imageManager.getImage(imageIndex),
                        0,
                        0,
                        this
                        );
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept mouse motion events

    public void processMouseMotionEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        switch (e.getID())
        {
            case MouseEvent.MOUSE_MOVED:
                break;

            case MouseEvent.MOUSE_DRAGGED:
                if (e.getX() < 0 || e.getY() < 0 || e.getX() > getBounds().width ||
                    e.getY() > getBounds().height)
                    mouseInside = false;
                else
                    mouseInside = true;
                repaint();
                break;
        }

        super.processMouseMotionEvent(e);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept mouse events

    public void processMouseEvent(MouseEvent e)
    {
        if (!isEnabled())
            return;

        switch (e.getID())
        {
            case MouseEvent.MOUSE_CLICKED:
                //MOUSE_RELEASED is much easier to satisfy for a "click"
                break;

            case MouseEvent.MOUSE_PRESSED:
                requestFocus();
                mouseDown = true;
                repaint();
                break;

            case MouseEvent.MOUSE_RELEASED:
                mouseDown = false;
                if (actionListener != null && mouseInside)
                {
                    actionListener.actionPerformed(
                        new ActionEvent(
                            this,
                            ActionEvent.ACTION_PERFORMED,
                            ""
                            )
                        );
                }
                repaint();
                break;

            case MouseEvent.MOUSE_ENTERED:
                mouseInside = true;
                repaint();
                break;

            case MouseEvent.MOUSE_EXITED:
                mouseInside = false;
                repaint();
                break;
        }

        super.processMouseEvent(e);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept focus events

    public void processFocusEvent(FocusEvent e)
    {
        switch (e.getID())
        {
            case FocusEvent.FOCUS_GAINED:
                hasFocus = true;
                repaint();
                break;

            case FocusEvent.FOCUS_LOST:
                hasFocus = false;
                repaint();
                break;
        }

        super.processFocusEvent(e);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept key events

    public void processKeyEvent(KeyEvent e)
    {
        if (!isEnabled())
            return;

        if (e.getID() == KeyEvent.KEY_PRESSED)
        {
            switch(e.getKeyChar())
            {
                case ' ':
                case '\n':
                    if (actionListener != null)
                    {
                        actionListener.actionPerformed(
                            new ActionEvent(
                                this,
                                ActionEvent.ACTION_PERFORMED,
                                ""
                                )
                            );
                    }
                    repaint();
                    break;
            }
        }

        super.processKeyEvent(e);
    }
}