package toy.dialog;

import java.awt.*;
import java.awt.event.*;

import toy.*;

/**
 *
 * @author Brian Tsang
 * @version 7.0
 */

public class ToyTextAreaPeer extends Container
                             implements ActionListener, AdjustmentListener,
                                        FocusListener, Runnable

{
    public static final int SCROLLBAR_MARGIN = 1;
    public static final int NORMAL_MARGIN    = 4;
    public static final int LINE_MARGIN      = 0;

    public static final Color OUTER_STREAK         = new Color(128, 128, 128);
    public static final Color INNER_STREAK         = new Color(0, 0, 0);

    public static final int SCROLLBAR_WIDTH        = 16;

    public static final String FONT_STRING         = "Monospaced";
    public static final int FONT_SIZE              = 12;

    public static final Color SELECTION_COLOR      = new Color(64, 81, 96);
    public static final Color SELECTION_TEXT_COLOR = new Color(255, 255, 255);

    public static final Font  HEADER_FONT          = new Font(FONT_STRING, Font.PLAIN, FONT_SIZE);
    public static final Color HEADER_COLOR         = new Color(0, 0, 0);

    public static final Font  KEYWORD_FONT         = new Font(FONT_STRING, Font.BOLD, FONT_SIZE);
    public static final Color KEYWORD_COLOR        = new Color(0, 0, 0);

    public static final Font  COMMAND_FONT         = new Font(FONT_STRING, Font.PLAIN, FONT_SIZE);
    public static final Color COMMAND_COLOR        = new Color(0, 0, 255);

    public static final Font  AUTOCOMMENT_FONT     = new Font(FONT_STRING, Font.PLAIN, FONT_SIZE);
    public static final Color AUTOCOMMENT_COLOR    = new Color(0, 0, 0);

    public static final Font  USERCOMMENT_FONT     = new Font(FONT_STRING, Font.PLAIN, FONT_SIZE);
    public static final Color USERCOMMENT_COLOR    = new Color(0, 128, 0);

    public static final Font  OUTOFBOUNDS_FONT     = new Font(FONT_STRING, Font.PLAIN, FONT_SIZE);
    public static final Color OUTOFBOUNDS_COLOR    = new Color(255, 0, 0);

    private Scrollbar verticalScrollbar;
    private Scrollbar horizontalScrollbar;
    private Panel greyPanel;

    private ToyTextArea parentTextArea;

    private boolean hasFocus;

    private int totalWidth;
    private int totalHeight;

    private String currentText;
    private String wrappedText[];
    private int barLine;

    private Image offscreenImage;
    private Graphics offscreen;

    private Thread runner;
    private boolean caretVisible;
    private long lastCaretRepaint;

    private int selectionStartPosition;

    private int fontWidth;
    private int fontHeight;

    public ToyTextAreaPeer(ToyTextArea newParentTextArea)
    {
        super();

        currentText = "";
        wrappedText = new String[1];
        wrappedText[0] = "";

        parentTextArea = newParentTextArea;
        parentTextArea.addActionListener(this);
        parentTextArea.addFocusListener(this);

        horizontalScrollbar = new Scrollbar(Scrollbar.HORIZONTAL);
        horizontalScrollbar.addAdjustmentListener(this);
        horizontalScrollbar.setVisible(false);
        add(horizontalScrollbar);

        verticalScrollbar = new Scrollbar();
        verticalScrollbar.addAdjustmentListener(this);
        add(verticalScrollbar);

        greyPanel = new Panel();
        greyPanel.setBackground(new Color(192, 192, 192));
        add(greyPanel);

        FontMetrics fontMetrics = ToyToolkit.getFontMetrics(HEADER_FONT);

        fontHeight = fontMetrics.getHeight();
        fontWidth = fontMetrics.charWidth('0');

        enableEvents(AWTEvent.MOUSE_EVENT_MASK | AWTEvent.MOUSE_MOTION_EVENT_MASK);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Describe our new Component

    public boolean isFocusTraversable()
    {
        return false;
    }

    public boolean isDoubleBuffered()
    {
        return true;
    }

    public boolean isLightweight()
    {
        return true;
    }

    public Dimension getMinimumSize()
    {
        return parentTextArea.getMinimumSize();
    }

    public Dimension getPreferredSize()
    {
        return parentTextArea.getPreferredSize();
    }

    public Dimension getMaximumSize()
    {
        return parentTextArea.getMaximumSize();
    }

    ////////////////////////////////////////////////////////////////////////////
    // intercept functions
    public void setBounds(Rectangle r)
    {
        setBounds(r.x, r.y, r.width, r.height);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        super.setBounds(x, y, width, height);

        offscreen = null;
        refresh();
    }

    ////////////////////////////////////////////////////////////////////////////
    // special functions
    private int getVisibleWidth()
    {
        return getBounds().width - SCROLLBAR_MARGIN - NORMAL_MARGIN - SCROLLBAR_WIDTH;
    }

    private int getVisibleHeight()
    {
        if (needsHorizontalScrollbar())
            return getBounds().height - SCROLLBAR_MARGIN - NORMAL_MARGIN - SCROLLBAR_WIDTH;
        else
            return getBounds().height - 2 * NORMAL_MARGIN;
    }

    private boolean needsHorizontalScrollbar()
    {
        return getVisibleWidth() < totalWidth;
    }

    private Point getCaretCoordinate()
    {
        return getPositionCoordinate(parentTextArea.getCaretPosition());
    }

    private Point getPositionCoordinate(int position)
    {
        if (position <= 0)
            position = 0;
        if (position >= currentText.length())
            position = currentText.length();

        int x = 0;
        int y = 0;

        for (int ctr = 0; ctr < position; ctr++)
        {
            x++;

            if (currentText.charAt(ctr) == '\n')
            {
                y++;
                x = 0;
            }
        }

        return new Point(x, y);
    }

    private Point getPixelCoordinate(Point pixel)
    {
        int x, y;

        x = (pixel.x - NORMAL_MARGIN + horizontalScrollbar.getValue()) / fontWidth;
        y = (pixel.y - NORMAL_MARGIN + verticalScrollbar.getValue()) / (fontHeight + LINE_MARGIN);

        if (y < 0)
            y = 0;
        if (y >= wrappedText.length)
            y = wrappedText.length - 1;

        if (x < 0)
            x = 0;
        if (x > wrappedText[y].length())
            x = wrappedText[y].length();

        return new Point(x, y);
    }

    private int getCoordinatePosition(Point p)
    {
        return getCoordinatePosition(p.x, p.y);
    }

    private int getCoordinatePosition(int x, int y)
    {
        int answer = 0;

        for (int ctr = 0; ctr < y && ctr < wrappedText.length; ctr++)
            answer += wrappedText[ctr].length() + 1; //the newline

        if (y < wrappedText.length)
            answer += Math.min(wrappedText[y].length(), x);

        return answer;
    }

    public void refresh()
    {
        if (!isVisible())
            return;

        //recalculate wrapped text and height
        if (!currentText.equals(parentTextArea.getText(true)))
        {
            currentText = parentTextArea.getText(true);

            //count the number of newlines
            int lines = 1;
            for (int ctr = 0; ctr < currentText.length(); ctr++)
                if (currentText.charAt(ctr) == '\n')
                    lines++;

            //ok now start dumping the lines into an array
            wrappedText = new String[lines];
            int lastNewline = -1;
            for (int ctr = 0; ctr < lines - 1; ctr++)
            {
                int nextNewline = currentText.indexOf('\n', lastNewline + 1);

                wrappedText[ctr] = currentText.substring(lastNewline + 1, nextNewline);

                lastNewline = nextNewline;
            }
            wrappedText[lines - 1] = currentText.substring(lastNewline + 1);

            //recalculate barline, height and width
            totalHeight = (fontHeight + LINE_MARGIN) * wrappedText.length
                - LINE_MARGIN + 3;

            totalWidth = wrappedText[0].length() * fontWidth;
            barLine = -1;
            for (int ctr = 1; ctr < wrappedText.length; ctr++)
            {
                if (barLine == -1 && wrappedText[ctr].equals(ToyWorkspace.HEADER_BAR))
                    barLine = ctr;
                totalWidth = Math.max(totalWidth, wrappedText[ctr].length() * fontWidth);
            }
        }

        //resize components
        if (needsHorizontalScrollbar())
        {
            horizontalScrollbar.setVisible(true);
            greyPanel.setVisible(true);

            horizontalScrollbar.setBounds(
                0,
                getBounds().height - SCROLLBAR_WIDTH,
                getBounds().width - SCROLLBAR_WIDTH,
                SCROLLBAR_WIDTH
                );

            horizontalScrollbar.setMinimum(0);
            horizontalScrollbar.setMaximum(totalWidth);
            horizontalScrollbar.setVisibleAmount(getVisibleWidth());

            verticalScrollbar.setBounds(
                getBounds().width - SCROLLBAR_WIDTH,
                0,
                SCROLLBAR_WIDTH,
                getBounds().height - SCROLLBAR_WIDTH
                );
            verticalScrollbar.setMinimum(3);
            verticalScrollbar.setMaximum(totalHeight);
            verticalScrollbar.setVisibleAmount(getVisibleHeight());

            greyPanel.setBounds(
                getBounds().width - SCROLLBAR_WIDTH,
                getBounds().height - SCROLLBAR_WIDTH,
                SCROLLBAR_WIDTH,
                SCROLLBAR_WIDTH
                );
        }
        else {
            horizontalScrollbar.setVisible(false);
            horizontalScrollbar.setValue(0);
            greyPanel.setVisible(false);

            verticalScrollbar.setBounds(
                getBounds().width - SCROLLBAR_WIDTH,
                0,
                SCROLLBAR_WIDTH,
                getBounds().height
                );

            verticalScrollbar.setMinimum(3);
            verticalScrollbar.setMaximum(totalHeight);
            verticalScrollbar.setVisibleAmount(getVisibleHeight());
        }


        //check to make sure that the caret is on screen
        {
            Point caret = getCaretCoordinate();

            int caretX = NORMAL_MARGIN - horizontalScrollbar.getValue() +
                    fontWidth * caret.x;
            int caretY = NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                    (fontHeight + LINE_MARGIN) * caret.y - 3;

            if (caretX < NORMAL_MARGIN)
                horizontalScrollbar.setValue(fontWidth * caret.x - 10);
            if (caretX > getBounds().width - SCROLLBAR_MARGIN - SCROLLBAR_WIDTH)
                horizontalScrollbar.setValue(
                    -getBounds().width + NORMAL_MARGIN + SCROLLBAR_MARGIN +
                        SCROLLBAR_WIDTH + fontWidth * caret.x + 10
                    );
            if (caretY + 3 - fontHeight < NORMAL_MARGIN)
                verticalScrollbar.setValue(
                    (fontHeight + LINE_MARGIN) * caret.y - 7
                    );

            if (needsHorizontalScrollbar())
            {
                if (caretY + 3 > getBounds().height - SCROLLBAR_MARGIN - SCROLLBAR_WIDTH)
                    verticalScrollbar.setValue(
                        -getBounds().height + NORMAL_MARGIN + SCROLLBAR_MARGIN +
                            SCROLLBAR_WIDTH + fontHeight +
                            (fontHeight + LINE_MARGIN) * caret.y + 13
                        );
            }
            else {
                if (caretY + 3 > getBounds().height - NORMAL_MARGIN)
                    verticalScrollbar.setValue(
                        -getBounds().height + 2 * NORMAL_MARGIN + fontHeight +
                            (fontHeight + LINE_MARGIN) * caret.y + 13
                        );
            }
        }

        caretVisible = hasFocus;
        lastCaretRepaint = System.currentTimeMillis();
        repaint();
    }

    ////////////////////////////////////////////////////////////////////////////
    // Overriding the painting functions
    public void repaint()
    {
        if (getParent() == null || !isVisible())
            return;

        paint(getGraphics());
    }

    public void repaint(int x, int y, int width, int height)
    {
        repaint();
    }

    public void repaint(long tm)
    {

    }

    public void update(Graphics g)
    {
        if (getParent() == null || !isVisible())
            return;

        paint(g);
    }

    public void paint(Graphics g)
    {
        if (getParent() == null || !isVisible() || g == null ||
            getBounds().width <= 0 || getBounds().height <= 0)
            return;

        if (offscreen == null)
        {
            offscreenImage = createImage(getBounds().width, getBounds().height);

            if (offscreenImage == null)
                return;

            offscreen = offscreenImage.getGraphics();

            if (offscreen == null)
                return;
        }

        try
        {
            setBackground(parentTextArea.getBackground());
            setForeground(parentTextArea.getForeground());

            float backgroundHSB[] = Color.RGBtoHSB(
                getBackground().getRed(),
                getBackground().getGreen(),
                getBackground().getBlue(),
                null
                );

            Point selectionStart = getPositionCoordinate(
                Math.min(
                    parentTextArea.getSelectionStart(),
                    parentTextArea.getSelectionEnd()
                    )
                );

            Point selectionEnd = getPositionCoordinate(
                Math.max(
                    parentTextArea.getSelectionStart(),
                    parentTextArea.getSelectionEnd()
                    )
                );

            offscreen.setClip(0, 0, getBounds().width, getBounds().height);

            //white filling
            offscreen.setColor(getBackground());
            if (needsHorizontalScrollbar())
                offscreen.fillRect(0, 0, getBounds().width - SCROLLBAR_WIDTH, getBounds().height - SCROLLBAR_WIDTH);
            else
                offscreen.fillRect(0, 0, getBounds().width - SCROLLBAR_WIDTH, getBounds().height);

            //top and left borders
            offscreen.setColor(
                Color.getHSBColor(
                    backgroundHSB[0],
                    backgroundHSB[1],
                    .5F * backgroundHSB[2]
                    )
                );
            offscreen.drawLine(0, 0, getBounds().width, 0);
            offscreen.drawLine(0, 0, 0, getBounds().height);
            offscreen.setColor(new Color(0, 0, 0));
            offscreen.drawLine(1, 1, getBounds().width, 1);
            offscreen.drawLine(1, 1, 1, getBounds().height);

            //bottom border (if there's no horizontal scrollbar)
            if (!needsHorizontalScrollbar())
            {
                offscreen.setColor(
                    Color.getHSBColor(
                        backgroundHSB[0],
                        backgroundHSB[1],
                        .75F * backgroundHSB[2]
                        )
                    );
                offscreen.drawLine(
                    1,
                    getBounds().height - 2,
                    getBounds().width,
                    getBounds().height - 2
                    );
                offscreen.setColor(
                    Color.getHSBColor(
                        backgroundHSB[0],
                        backgroundHSB[1],
                        .9F * backgroundHSB[2]
                        )
                    );
                offscreen.drawLine(
                    0,
                    getBounds().height - 1,
                    getBounds().width,
                    getBounds().height - 1
                    );
            }

            //draw the current line highlight
            if (needsHorizontalScrollbar())
                offscreen.setClip(
                    2,
                    2,
                    getBounds().width - SCROLLBAR_WIDTH,
                    getBounds().height - SCROLLBAR_WIDTH
                    );
            else
                offscreen.setClip(
                    2,
                    2,
                    getBounds().width - SCROLLBAR_WIDTH,
                    getBounds().height - 2
                    );

            if (hasFocus)
            {
                offscreen.setColor(
                    Color.getHSBColor(
                        backgroundHSB[0],
                        backgroundHSB[1],
                        .875F * backgroundHSB[2]
                        )
                    );

                offscreen.fillRect(
                    0,
                    NORMAL_MARGIN - verticalScrollbar.getValue() +
                        (fontHeight + LINE_MARGIN) * getCaretCoordinate().y + 3,
                    getBounds().width,
                    fontHeight + LINE_MARGIN
                );
            }

            //draw the 41 and 80 character mark
            offscreen.setColor(
                Color.getHSBColor(
                    backgroundHSB[0],
                    backgroundHSB[1],
                    .75F * backgroundHSB[2]
                    )
                );
            offscreen.drawLine(
                NORMAL_MARGIN + fontWidth * (8 + ToyWorkspace.COMMENT_WIDTH)
                    - horizontalScrollbar.getValue(),
                NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight / 2 +
                        (fontHeight + LINE_MARGIN) * barLine,
                NORMAL_MARGIN + fontWidth * (8 + ToyWorkspace.COMMENT_WIDTH)
                    - horizontalScrollbar.getValue(),
                getBounds().height
                );
            offscreen.drawLine(
                NORMAL_MARGIN + fontWidth * 80 - horizontalScrollbar.getValue(),
                0,
                NORMAL_MARGIN + fontWidth * 80 - horizontalScrollbar.getValue(),
                getBounds().height
                );

            //color the background of the selected text
            if (parentTextArea.getSelectionStart() != parentTextArea.getSelectionEnd())
            {
                offscreen.setColor(SELECTION_COLOR);

                if (selectionStart.y == selectionEnd.y)
                    offscreen.fillRect(
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * selectionStart.x,
                        NORMAL_MARGIN - verticalScrollbar.getValue() +
                            (fontHeight + LINE_MARGIN) * selectionStart.y + 3,
                        fontWidth * (selectionEnd.x - selectionStart.x),
                        fontHeight + LINE_MARGIN
                    );
                else {
                    offscreen.fillRect(
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * selectionStart.x,
                        NORMAL_MARGIN - verticalScrollbar.getValue() +
                            (fontHeight + LINE_MARGIN) * selectionStart.y + 3,
                        getBounds().width - NORMAL_MARGIN
                            + horizontalScrollbar.getValue()
                            - fontWidth * selectionStart.x,
                        fontHeight + LINE_MARGIN
                    );

                    offscreen.fillRect(
                        NORMAL_MARGIN - horizontalScrollbar.getValue() - 1,
                        NORMAL_MARGIN - verticalScrollbar.getValue() +
                            (fontHeight + LINE_MARGIN) *
                                (selectionStart.y + 1) + 3,
                        getBounds().width - NORMAL_MARGIN
                            + horizontalScrollbar.getValue(),
                        (fontHeight + LINE_MARGIN) *
                            (selectionEnd.y - selectionStart.y - 1)
                        );

                    offscreen.fillRect(
                        NORMAL_MARGIN - horizontalScrollbar.getValue(),
                        NORMAL_MARGIN - verticalScrollbar.getValue() +
                            (fontHeight + LINE_MARGIN) * selectionEnd.y + 3,
                        fontWidth * selectionEnd.x,
                        fontHeight + LINE_MARGIN
                    );
                }
            }

            int currentLine = getPixelCoordinate(new Point (0, 0)).y;
            int endLine = getPixelCoordinate(new Point(0, getBounds().height)).y;

            //draw the header
            if (needsHorizontalScrollbar())
                offscreen.setClip(
                    NORMAL_MARGIN,
                    NORMAL_MARGIN,
                    getBounds().width - NORMAL_MARGIN - SCROLLBAR_MARGIN - SCROLLBAR_WIDTH,
                    getBounds().height - NORMAL_MARGIN - SCROLLBAR_MARGIN - SCROLLBAR_WIDTH
                    );
            else
                offscreen.setClip(
                    NORMAL_MARGIN,
                    NORMAL_MARGIN,
                    getBounds().width - NORMAL_MARGIN - SCROLLBAR_MARGIN - SCROLLBAR_WIDTH,
                    getBounds().height - 2 * NORMAL_MARGIN
                    );

            while (currentLine < barLine && currentLine <= endLine)
            {
                int resumePosition = 0;

                //make "Program" bold but only on the first line
                if (currentLine == 0 && wrappedText[currentLine].startsWith("Program "))
                {
                    offscreen.setFont(KEYWORD_FONT);
                    offscreen.setColor(KEYWORD_COLOR);

                    offscreen.drawString(
                        "Program",
                        NORMAL_MARGIN - horizontalScrollbar.getValue(),
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );

                    int subSelectionStart = 0;
                    int subSelectionEnd = "Program".length();

                    if (selectionStart.y > currentLine)
                        subSelectionStart = wrappedText[currentLine].length();
                    if (selectionStart.y == currentLine)
                        subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                    if (selectionEnd.y < currentLine)
                        subSelectionEnd = 0;
                    if (selectionEnd.y == currentLine)
                        subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                    if (subSelectionStart < subSelectionEnd)
                    {
                        offscreen.setColor(SELECTION_TEXT_COLOR);

                        offscreen.drawString(
                            "Program".substring(subSelectionStart, subSelectionEnd),
                            NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                fontWidth * subSelectionStart,
                            NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                (fontHeight + LINE_MARGIN) * currentLine
                            );
                    }

                    resumePosition = "Program ".length();
                }

                //make Input, Output, and Remarks bold
                if (wrappedText[currentLine].startsWith("Input: "))
                {
                    offscreen.setFont(KEYWORD_FONT);
                    offscreen.setColor(KEYWORD_COLOR);

                    offscreen.drawString(
                        "Input",
                        NORMAL_MARGIN - horizontalScrollbar.getValue(),
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );

                    int subSelectionStart = 0;
                    int subSelectionEnd = "Input".length();

                    if (selectionStart.y > currentLine)
                        subSelectionStart = wrappedText[currentLine].length();
                    if (selectionStart.y == currentLine)
                        subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                    if (selectionEnd.y < currentLine)
                        subSelectionEnd = 0;
                    if (selectionEnd.y == currentLine)
                        subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                    if (subSelectionStart < subSelectionEnd)
                    {
                        offscreen.setColor(SELECTION_TEXT_COLOR);

                        offscreen.drawString(
                            "Input".substring(subSelectionStart, subSelectionEnd),
                            NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                fontWidth * subSelectionStart,
                            NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                (fontHeight + LINE_MARGIN) * currentLine
                            );
                    }

                    resumePosition = "Input".length();
                }

                if (wrappedText[currentLine].startsWith("Output: "))
                {
                    offscreen.setFont(KEYWORD_FONT);
                    offscreen.setColor(KEYWORD_COLOR);

                    offscreen.drawString(
                        "Output",
                        NORMAL_MARGIN - horizontalScrollbar.getValue(),
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );

                    int subSelectionStart = 0;
                    int subSelectionEnd = "Output".length();

                    if (selectionStart.y > currentLine)
                        subSelectionStart = wrappedText[currentLine].length();
                    if (selectionStart.y == currentLine)
                        subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                    if (selectionEnd.y < currentLine)
                        subSelectionEnd = 0;
                    if (selectionEnd.y == currentLine)
                        subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                    if (subSelectionStart < subSelectionEnd)
                    {
                        offscreen.setColor(SELECTION_TEXT_COLOR);

                        offscreen.drawString(
                            "Output".substring(subSelectionStart, subSelectionEnd),
                            NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                fontWidth * subSelectionStart,
                            NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                (fontHeight + LINE_MARGIN) * currentLine
                            );
                    }

                    resumePosition = "Output".length();
                }

                if (wrappedText[currentLine].startsWith("Remarks: "))
                {
                    offscreen.setFont(KEYWORD_FONT);
                    offscreen.setColor(KEYWORD_COLOR);

                    offscreen.drawString(
                        "Remarks",
                        NORMAL_MARGIN - horizontalScrollbar.getValue(),
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );

                    int subSelectionStart = 0;
                    int subSelectionEnd = "Remarks".length();

                    if (selectionStart.y > currentLine)
                        subSelectionStart = wrappedText[currentLine].length();
                    if (selectionStart.y == currentLine)
                        subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                    if (selectionEnd.y < currentLine)
                        subSelectionEnd = 0;
                    if (selectionEnd.y == currentLine)
                        subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                    if (subSelectionStart < subSelectionEnd)
                    {
                        offscreen.setColor(SELECTION_TEXT_COLOR);

                        offscreen.drawString(
                            "Remarks".substring(subSelectionStart, subSelectionEnd),
                            NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                fontWidth * subSelectionStart,
                            NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                (fontHeight + LINE_MARGIN) * currentLine
                            );
                    }

                    resumePosition = "Remarks".length();
                }

                if (wrappedText[currentLine].length() > 80)
                {
                    offscreen.setFont(HEADER_FONT);
                    offscreen.setColor(HEADER_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(resumePosition, 80),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * resumePosition,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );

                    offscreen.setFont(OUTOFBOUNDS_FONT);
                    offscreen.setColor(OUTOFBOUNDS_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(80),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * 80,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }
                else {
                    offscreen.setFont(HEADER_FONT);
                    offscreen.setColor(HEADER_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(resumePosition),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * resumePosition,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }

                int subSelectionStart = resumePosition;
                int subSelectionEnd = wrappedText[currentLine].length();

                if (selectionStart.y > currentLine)
                    subSelectionStart = wrappedText[currentLine].length();
                if (selectionStart.y == currentLine)
                    subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                if (selectionEnd.y < currentLine)
                    subSelectionEnd = 0;
                if (selectionEnd.y == currentLine)
                    subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                if (subSelectionStart < subSelectionEnd)
                {
                    offscreen.setColor(SELECTION_TEXT_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(subSelectionStart, subSelectionEnd),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * subSelectionStart,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }

                currentLine++;
            }

            //draw the bar
            if (currentLine == barLine && currentLine <= endLine)
            {
                offscreen.setColor(HEADER_COLOR);

                offscreen.drawLine(
                    NORMAL_MARGIN - horizontalScrollbar.getValue(),
                    NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight / 2 +
                        (fontHeight + LINE_MARGIN) * currentLine,
                    NORMAL_MARGIN - horizontalScrollbar.getValue() +
                        fontWidth * ToyWorkspace.HEADER_BAR.length() - 1,
                    NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight / 2 +
                        (fontHeight + LINE_MARGIN) * currentLine
                    );

                int subSelectionStart = 0;
                int subSelectionEnd = wrappedText[currentLine].length();

                if (selectionStart.y > currentLine)
                    subSelectionStart = wrappedText[currentLine].length();
                if (selectionStart.y == currentLine)
                    subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                if (selectionEnd.y < currentLine)
                    subSelectionEnd = 0;
                if (selectionEnd.y == currentLine)
                    subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                if (subSelectionStart < subSelectionEnd)
                {
                    offscreen.setColor(SELECTION_TEXT_COLOR);

                    offscreen.drawLine(
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * subSelectionStart,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight / 2 +
                            (fontHeight + LINE_MARGIN) * currentLine,
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * subSelectionEnd - 1,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight / 2 +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }

                currentLine++;
            }

            //draw the body
            while (currentLine <= endLine)
            {
                int resumePosition = 0;
                int charCtr = 0;

                //check to see if we have a command line
                while (charCtr + 7 < wrappedText[currentLine].length() &&
                       Character.isWhitespace(wrappedText[currentLine].charAt(charCtr)))
                    charCtr++;

                if (charCtr + 7 < wrappedText[currentLine].length() &&
                    ToyWord.isHexDigit(wrappedText[currentLine].charAt(charCtr)) &&
                    ToyWord.isHexDigit(wrappedText[currentLine].charAt(charCtr + 1)) &&
                    wrappedText[currentLine].charAt(charCtr + 2) == ':' &&
                    wrappedText[currentLine].charAt(charCtr + 3) == ' ' &&
                    ToyWord.isHexDigit(wrappedText[currentLine].charAt(charCtr + 4)) &&
                    ToyWord.isHexDigit(wrappedText[currentLine].charAt(charCtr + 5)) &&
                    ToyWord.isHexDigit(wrappedText[currentLine].charAt(charCtr + 6)) &&
                    ToyWord.isHexDigit(wrappedText[currentLine].charAt(charCtr + 7))
                    )
                {
                    resumePosition = Math.min(charCtr + 8, 80);
                    offscreen.setFont(COMMAND_FONT);
                    offscreen.setColor(COMMAND_COLOR);

                    if (charCtr < 80)
                    {
                        offscreen.drawString(
                            wrappedText[currentLine].substring(charCtr, resumePosition),
                            NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                fontWidth * charCtr,
                            NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                (fontHeight + LINE_MARGIN) * currentLine
                            );

                        int subSelectionStart = charCtr;
                        int subSelectionEnd = resumePosition;

                        if (selectionStart.y > currentLine)
                            subSelectionStart = wrappedText[currentLine].length();
                        if (selectionStart.y == currentLine)
                            subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                        if (selectionEnd.y < currentLine)
                            subSelectionEnd = 0;
                        if (selectionEnd.y == currentLine)
                            subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                        if (subSelectionStart < subSelectionEnd)
                        {
                            offscreen.setColor(SELECTION_TEXT_COLOR);

                            offscreen.drawString(
                                wrappedText[currentLine].substring(subSelectionStart, subSelectionEnd),
                                NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                    fontWidth * subSelectionStart,
                                NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                    (fontHeight + LINE_MARGIN) * currentLine
                                );
                        }

                        if (wrappedText[currentLine].length() > charCtr + 8)
                        {
                            resumePosition = Math.min(
                                wrappedText[currentLine].length(),
                                Math.min(charCtr + 8 + ToyWorkspace.COMMENT_WIDTH, 80)
                                );
                            offscreen.setFont(AUTOCOMMENT_FONT);
                            offscreen.setColor(AUTOCOMMENT_COLOR);

                            if (charCtr + 8 < 80)
                            {
                                offscreen.drawString(
                                    wrappedText[currentLine].substring(charCtr + 8, resumePosition),
                                    NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                        fontWidth * (charCtr + 8),
                                    NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                        (fontHeight + LINE_MARGIN) * currentLine
                                    );

                                subSelectionStart = charCtr + 8;
                                subSelectionEnd = resumePosition;

                                if (selectionStart.y > currentLine)
                                    subSelectionStart = wrappedText[currentLine].length();
                                if (selectionStart.y == currentLine)
                                    subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                                if (selectionEnd.y < currentLine)
                                    subSelectionEnd = 0;
                                if (selectionEnd.y == currentLine)
                                    subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                                if (subSelectionStart < subSelectionEnd)
                                {
                                    offscreen.setColor(SELECTION_TEXT_COLOR);

                                    offscreen.drawString(
                                        wrappedText[currentLine].substring(subSelectionStart, subSelectionEnd),
                                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                                            fontWidth * subSelectionStart,
                                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                                            (fontHeight + LINE_MARGIN) * currentLine
                                        );
                                }
                            }
                        }
                    }
                }


                if (wrappedText[currentLine].length() > 80)
                {
                    offscreen.setFont(USERCOMMENT_FONT);
                    offscreen.setColor(USERCOMMENT_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(resumePosition, 80),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * resumePosition,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );

                    offscreen.setFont(OUTOFBOUNDS_FONT);
                    offscreen.setColor(OUTOFBOUNDS_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(80),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * 80,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }
                else {
                    offscreen.setFont(USERCOMMENT_FONT);
                    offscreen.setColor(USERCOMMENT_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(resumePosition),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * resumePosition,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }

                int subSelectionStart = resumePosition;
                int subSelectionEnd = wrappedText[currentLine].length();

                if (selectionStart.y > currentLine)
                    subSelectionStart = wrappedText[currentLine].length();
                if (selectionStart.y == currentLine)
                    subSelectionStart = Math.max(subSelectionStart, selectionStart.x);

                if (selectionEnd.y < currentLine)
                    subSelectionEnd = 0;
                if (selectionEnd.y == currentLine)
                    subSelectionEnd = Math.min(subSelectionEnd, selectionEnd.x);

                if (subSelectionStart < subSelectionEnd)
                {
                    offscreen.setColor(SELECTION_TEXT_COLOR);

                    offscreen.drawString(
                        wrappedText[currentLine].substring(subSelectionStart, subSelectionEnd),
                        NORMAL_MARGIN - horizontalScrollbar.getValue() +
                            fontWidth * subSelectionStart,
                        NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                            (fontHeight + LINE_MARGIN) * currentLine
                        );
                }

                currentLine++;
            }

            //draw the caret
            if (caretVisible)
            {
                Point caret = getCaretCoordinate();

                offscreen.setColor(new Color(0, 0, 0));
                offscreen.drawLine(
                    NORMAL_MARGIN - horizontalScrollbar.getValue() +
                        fontWidth * caret.x,
                    NORMAL_MARGIN - verticalScrollbar.getValue() + fontHeight +
                        (fontHeight + LINE_MARGIN) * caret.y + 3,
                    NORMAL_MARGIN - horizontalScrollbar.getValue() +
                        fontWidth * caret.x,
                    NORMAL_MARGIN - verticalScrollbar.getValue() +
                        (fontHeight + LINE_MARGIN) * caret.y + 1
                    );
            }

            //draw the components
            for(int ctr = 0; ctr < getComponentCount(); ctr++)
            {
                offscreen.setClip(getComponent(ctr).getBounds());

                offscreen.translate(
                    getComponent(ctr).getBounds().x,
                    getComponent(ctr).getBounds().y
                    );

                getComponent(ctr).paint(offscreen);

                offscreen.translate(
                    -getComponent(ctr).getBounds().x,
                    -getComponent(ctr).getBounds().y
                    );
            }

            g.drawImage(offscreenImage, 0, 0, this);
        }
        catch (Exception e)
        {
            //whatever
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Accept mouse events

    public void processMouseEvent(MouseEvent e)
    {
        if (e.getID() == MouseEvent.MOUSE_ENTERED || e.getID() == MouseEvent.MOUSE_EXITED)
        {
            if (greyPanel.isVisible() && greyPanel.getBounds().contains(e.getX(), e.getY()) ||
                horizontalScrollbar.isVisible() && horizontalScrollbar.getBounds().contains(e.getX(), e.getY()) ||
                verticalScrollbar.isVisible() && verticalScrollbar.getBounds().contains(e.getX(), e.getY()))
            {
                if (getCursor() != Cursor.getDefaultCursor())
                    setCursor(Cursor.getDefaultCursor());
            }
            else {
                if (getCursor() != Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR))
                    setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
            }
        }

        if (e.getID() == MouseEvent.MOUSE_PRESSED)
        {
            if (!hasFocus)
                parentTextArea.requestFocus();

            selectionStartPosition = getCoordinatePosition(getPixelCoordinate(e.getPoint()));

            parentTextArea.setCaretPosition(selectionStartPosition);
        }

    }

    public void processMouseMotionEvent(MouseEvent e)
    {
        if (e.getID() == MouseEvent.MOUSE_MOVED)
        {
            if (greyPanel.isVisible() && greyPanel.getBounds().contains(e.getX(), e.getY()) ||
                horizontalScrollbar.isVisible() && horizontalScrollbar.getBounds().contains(e.getX(), e.getY()) ||
                verticalScrollbar.isVisible() && verticalScrollbar.getBounds().contains(e.getX(), e.getY()))
            {
                if (getCursor() != Cursor.getDefaultCursor())
                    setCursor(Cursor.getDefaultCursor());
            }
            else {
                if (getCursor() != Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR))
                    setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
            }
        }

        if (e.getID() == MouseEvent.MOUSE_DRAGGED)
        {
            if (!hasFocus)
                parentTextArea.requestFocus();

            int selectionEndPosition = getCoordinatePosition(getPixelCoordinate(e.getPoint()));

            parentTextArea.select(selectionStartPosition, selectionEndPosition);
        }

    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement Runnable
    public void run()
    {
        Thread thisThread = Thread.currentThread();
        while (thisThread == runner)
        {
            try
            {
                caretVisible = !caretVisible;
                lastCaretRepaint = System.currentTimeMillis();
                repaint();

                while (thisThread == runner &&
                       System.currentTimeMillis() - lastCaretRepaint < 500)
                    Thread.sleep(250);
            }
            catch (Exception e)
            {
                //whatever
            }
        }

        caretVisible = false;
        repaint();
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement ActionListener
    public void actionPerformed(ActionEvent e)
    {
        if (e.getSource() == parentTextArea && isVisible())
        {
            if (e.getActionCommand().equals("repaint()"))
                repaint();
            else
            {
                if (e.getActionCommand().equals("pageUp()") ||
                    e.getActionCommand().equals("pageDown()") ||
                    e.getActionCommand().equals("pageUpSelect()") ||
                    e.getActionCommand().equals("pageDownSelect()"))
                {
                    Point caret;

                    caret = getCaretCoordinate();

                    if (e.getActionCommand().equals("pageUp()") ||
                        e.getActionCommand().equals("pageUpSelect()"))
                        caret.y = Math.max(
                            0,
                            caret.y - getVisibleHeight() / fontHeight
                            );
                    else
                        caret.y = Math.min(
                            wrappedText.length - 1,
                            caret.y + getVisibleHeight() / fontHeight
                            );

                    caret.x = Math.min(
                        wrappedText[caret.y].length(),
                        caret.x
                        );

                    int caretPosition = getCoordinatePosition(caret);

                    if (e.getActionCommand().equals("pageUp()") ||
                        e.getActionCommand().equals("pageDown()"))
                        parentTextArea.setCaretPosition(caretPosition);
                    else {
                        int start = parentTextArea.getSelectionStart();
                        int end = parentTextArea.getSelectionEnd();

                        if (parentTextArea.getCaretPosition() == end)
                            parentTextArea.select(start, caretPosition);
                        else
                            parentTextArea.select(end, caretPosition);
                    }
                }

                refresh();
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement AdjustmentListener
    public void adjustmentValueChanged(AdjustmentEvent e)
    {
        if (e.getSource() == horizontalScrollbar)
            repaint();

        if (e.getSource() == verticalScrollbar)
            repaint();
    }

    ////////////////////////////////////////////////////////////////////////////
    // Implement FocusListener
    public void focusGained(FocusEvent e)
    {
        if (e.getSource() == parentTextArea)
        {
            hasFocus = true;

            if (runner == null)
            {
                runner = new Thread(this);
                runner.start();
            }
        }
    }

    public void focusLost(FocusEvent e)
    {
        if (e.getSource() == parentTextArea)
        {
            hasFocus = false;
            runner = null;
        }
    }
}